@@ -1,27 +1,25 @@
-// Copyright (C) 2016 Yixuan Qiu <yixuan.qiu@cos.name>
+// Copyright (C) 2016-2018 Yixuan Qiu <yixuan.qiu@cos.name>
 //
 // This Source Code Form is subject to the terms of the Mozilla
 // Public License v. 2.0. If a copy of the MPL was not distributed
-// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
+// with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
 
 #ifndef SYM_EIGS_SOLVER_H
 #define SYM_EIGS_SOLVER_H
 
 #include <Eigen/Core>
-#include <Eigen/Eigenvalues>
 #include <vector>     // std::vector
-#include <cmath>      // std::abs, std::pow
+#include <cmath>      // std::abs, std::pow, std::sqrt
 #include <algorithm>  // std::min, std::copy
-#include <limits>     // std::numeric_limits
 #include <stdexcept>  // std::invalid_argument
 
-#include "SelectionRule.h"
-#include "CompInfo.h"
-#include "SimpleRandom.h"
+#include "Util/TypeTraits.h"
+#include "Util/SelectionRule.h"
+#include "Util/CompInfo.h"
+#include "Util/SimpleRandom.h"
 #include "LinAlg/UpperHessenbergQR.h"
 #include "LinAlg/TridiagEigen.h"
 #include "MatOp/DenseSymMatProd.h"
-#include "MatOp/DenseSymShiftSolve.h"
 
 
 namespace Spectra {
@@ -36,16 +34,16 @@ namespace Spectra {
 ///
 /// \ingroup EigenSolver
 ///
-/// This class implements the eigen solver for real symmetric matrices.
+/// This class implements the eigen solver for real symmetric matrices, i.e.,
+/// to solve \f$Ax=\lambda x\f$ where \f$A\f$ is symmetric.
 ///
 /// **Spectra** is designed to calculate a specified number (\f$k\f$)
 /// of eigenvalues of a large square matrix (\f$A\f$). Usually \f$k\f$ is much
 /// less than the size of the matrix (\f$n\f$), so that only a few eigenvalues
 /// and eigenvectors are computed.
 ///
-/// This class implements the eigen solver of a real symmetric matrix, but
-/// rather than providing the whole matrix, the algorithm only requires the
-/// matrix-vector multiplication operation of \f$A\f$. Therefore, users of
+/// Rather than providing the whole \f$A\f$ matrix, the algorithm only requires
+/// the matrix-vector multiplication operation of \f$A\f$. Therefore, users of
 /// this solver need to supply a class that computes the result of \f$Av\f$
 /// for any given vector \f$v\f$. The name of this class should be given to
 /// the template parameter `OpType`, and instance of this class passed to
@@ -55,7 +53,8 @@ namespace Spectra {
 /// for example `Eigen::MatrixXd`, then there is an easy way to construct such
 /// matrix operation class, by using the built-in wrapper class DenseSymMatProd
 /// which wraps an existing matrix object in **Eigen**. This is also the
-/// default template parameter for SymEigsSolver.
+/// default template parameter for SymEigsSolver. For sparse matrices, the
+/// wrapper class SparseSymMatProd can be used similarly.
 ///
 /// If the users need to define their own matrix-vector multiplication operation
 /// class, it should implement all the public member functions as in DenseSymMatProd.
@@ -68,7 +67,8 @@ namespace Spectra {
 ///                       The full list of enumeration values can be found in
 ///                       \ref Enumerations.
 /// \tparam OpType        The name of the matrix operation class. Users could either
-///                       use the DenseSymMatProd wrapper class, or define their
+///                       use the wrapper classes such as DenseSymMatProd and
+///                       SparseSymMatProd, or define their
 ///                       own that impelemnts all the public member functions as in
 ///                       DenseSymMatProd.
 ///
@@ -87,7 +87,7 @@ namespace Spectra {
 ///     Eigen::MatrixXd A = Eigen::MatrixXd::Random(10, 10);
 ///     Eigen::MatrixXd M = A + A.transpose();
 ///
-///     // Construct matrix operation object using the wrapper class DenseGenMatProd
+///     // Construct matrix operation object using the wrapper class DenseSymMatProd
 ///     DenseSymMatProd<double> op(M);
 ///
 ///     // Construct eigen solver object, requesting the largest three eigenvalues
@@ -124,7 +124,7 @@ namespace Spectra {
 ///     int rows() { return 10; }
 ///     int cols() { return 10; }
 ///     // y_out = M * x_in
-///     void perform_op(double* x_in, double* y_out)
+///     void perform_op(double *x_in, double *y_out)
 ///     {
 ///         for(int i = 0; i < rows(); i++)
 ///         {
@@ -162,50 +162,71 @@ private:
     typedef Eigen::Array<bool, Eigen::Dynamic, 1> BoolArray;
     typedef Eigen::Map<Matrix> MapMat;
     typedef Eigen::Map<Vector> MapVec;
-    typedef Eigen::SelfAdjointEigenSolver<Matrix> EigenSolver;
 
 protected:
-    OpType *m_op;          // object to conduct matrix operation,
-                           // e.g. matrix-vector product
+    OpType*      m_op;         // object to conduct matrix operation,
+                               // e.g. matrix-vector product
+    const int    m_n;          // dimension of matrix A
+    const int    m_nev;        // number of eigenvalues requested
+    const int    m_ncv;        // dimension of Krylov subspace in the Lanczos method
+    int          m_nmatop;     // number of matrix operations called
+    int          m_niter;      // number of restarting iterations
+
+    Matrix       m_fac_V;      // V matrix in the Lanczos factorization
+    Matrix       m_fac_H;      // H matrix in the Lanczos factorization
+    Vector       m_fac_f;      // residual in the Lanczos factorization
+    Vector       m_ritz_val;   // Ritz values
 
 private:
-    const int m_n;         // dimension of matrix A
+    Matrix       m_ritz_vec;   // Ritz vectors
+    Vector       m_ritz_est;   // last row of m_ritz_vec, also called the Ritz estimates
+    BoolArray    m_ritz_conv;  // indicator of the convergence of Ritz values
+    int          m_info;       // status of the computation
 
-protected:
-    const int m_nev;       // number of eigenvalues requested
+    const Scalar m_near_0;     // a very small value, but 1.0 / m_near_0 does not overflow
+                               // ~= 1e-307 for the "double" type
+    const Scalar m_eps;        // the machine precision, ~= 1e-16 for the "double" type
+    const Scalar m_eps23;      // m_eps^(2/3), used to test the convergence
 
-private:
-    const int m_ncv;       // number of ritz values
-    int m_nmatop;          // number of matrix operations called
-    int m_niter;           // number of restarting iterations
-
-    Matrix m_fac_V;        // V matrix in the Arnoldi factorization
-    Matrix m_fac_H;        // H matrix in the Arnoldi factorization
-    Vector m_fac_f;        // residual in the Arnoldi factorization
-
-protected:
-    Vector m_ritz_val;     // ritz values
-
-private:
-    Matrix m_ritz_vec;     // ritz vectors
-    Vector m_ritz_est;     // last row of m_ritz_vec
-    BoolArray m_ritz_conv; // indicator of the convergence of ritz values
-    int m_info;            // status of the computation
-
-    const Scalar m_prec;   // precision parameter used to test convergence
-                           // m_prec = epsilon^(2/3)
-                           // epsilon is the machine precision,
-                           // e.g. ~= 1e-16 for the "double" type
-
-    // Arnoldi factorization starting from step-k
-    void factorize_from(int from_k, int to_m, const Vector &fk)
+    // Given orthonormal basis functions V, find a nonzero vector f such that V'f = 0
+    // Assume that f has been properly allocated
+    void expand_basis(const MapMat& V, const int seed, Vector& f, Scalar& fnorm)
     {
+        using std::sqrt;
+
+        const Scalar thresh = m_eps * sqrt(Scalar(m_n));
+        for(int iter = 0; iter < 5; iter++)
+        {
+            // Randomly generate a new vector and orthogonalize it against V
+            SimpleRandom<Scalar> rng(seed + 123 * iter);
+            f.noalias() = rng.random_vec(m_n);
+            // f <- f - V * V' * f, so that f is orthogonal to V
+            Vector Vf = inner_product(V, m_fac_f);
+            f -= V * Vf;
+            // fnorm <- ||f||
+            fnorm = m_fac_f.norm();
+
+            // If fnorm is too close to zero, we try a new random vector,
+            // otherwise return the result
+            if(fnorm >= thresh)
+                return;
+        }
+    }
+
+    // Lanczos factorization starting from step-k
+    void factorize_from(int from_k, int to_m, const Vector& fk)
+    {
+        using std::sqrt;
+
         if(to_m <= from_k) return;
 
-        m_fac_f = fk;
+        const Scalar beta_thresh = m_eps * sqrt(Scalar(m_n));
+        m_fac_f.noalias() = fk;
 
+        // Pre-allocate Vf
+        Vector Vf(to_m);
         Vector w(m_n);
-        Scalar beta = m_fac_f.norm(), Hii = 0.0;
+        Scalar beta = norm(m_fac_f), Hii = Scalar(0);
         // Keep the upperleft k x k submatrix of H and set other elements to 0
         m_fac_H.rightCols(m_ncv - from_k).setZero();
         m_fac_H.block(from_k, 0, m_ncv - from_k, from_k).setZero();
@@ -215,17 +236,10 @@ private:
             // If beta = 0, then the next V is not full rank
             // We need to generate a new residual vector that is orthogonal
             // to the current V, which we call a restart
-            if(beta < m_prec)
+            if(beta < m_near_0)
             {
-                SimpleRandom<Scalar> rng(2 * i);
-                m_fac_f.noalias() = rng.random_vec(m_n);
-                // f <- f - V * V' * f, so that f is orthogonal to V
                 MapMat V(m_fac_V.data(), m_n, i); // The first i columns
-                Vector Vf = V.transpose() * m_fac_f;
-                m_fac_f.noalias() -= V * Vf;
-                // beta <- ||f||
-                beta = m_fac_f.norm();
-
+                expand_basis(V, 2 * i, m_fac_f, beta);
                 restart = true;
             }
 
@@ -234,16 +248,13 @@ private:
             v.noalias() = m_fac_f / beta;
 
             // Note that H[i+1, i] equals to the unrestarted beta
-            if(restart)
-                m_fac_H(i, i - 1) = 0.0;
-            else
-                m_fac_H(i, i - 1) = beta;
+            m_fac_H(i, i - 1) = restart ? Scalar(0) : beta;
 
             // w <- A * v
             m_op->perform_op(v.data(), w.data());
             m_nmatop++;
 
-            Hii = v.dot(w);
+            Hii = inner_product(v, w);
             m_fac_H(i - 1, i) = m_fac_H(i, i - 1); // Due to symmetry
             m_fac_H(i, i) = Hii;
 
@@ -254,32 +265,47 @@ private:
             else
                 m_fac_f.noalias() = w - m_fac_H(i, i - 1) * m_fac_V.col(i - 1) - Hii * v;
 
-            beta = m_fac_f.norm();
+            beta = norm(m_fac_f);
 
             // f/||f|| is going to be the next column of V, so we need to test
             // whether V' * (f/||f||) ~= 0
-            MapMat V(m_fac_V.data(), m_n, i + 1); // The first (i+1) columns
-            Vector Vf = V.transpose() * m_fac_f;
+            const int i1 = i + 1;
+            MapMat V(m_fac_V.data(), m_n, i1); // The first (i+1) columns
+            Vf.head(i1).noalias() = inner_product(V, m_fac_f);
+            Scalar ortho_err = Vf.head(i1).cwiseAbs().maxCoeff();
             // If not, iteratively correct the residual
             int count = 0;
-            while(count < 5 && Vf.cwiseAbs().maxCoeff() > m_prec * beta)
+            while(count < 5 && ortho_err > m_eps * beta)
             {
+                // There is an edge case: when beta=||f|| is close to zero, f mostly consists
+                // of noises of rounding errors, so the test [ortho_err < eps * beta] is very
+                // likely to fail. In particular, if beta=0, then the test is ensured to fail.
+                // Hence when this happens, we force f to be zero, and then restart in the
+                // next iteration.
+                if(beta < beta_thresh)
+                {
+                    m_fac_f.setZero();
+                    beta = Scalar(0);
+                    break;
+                }
+
                 // f <- f - V * Vf
-                m_fac_f.noalias() -= V * Vf;
+                m_fac_f.noalias() -= V * Vf.head(i1);
                 // h <- h + Vf
                 m_fac_H(i - 1, i) += Vf[i - 1];
                 m_fac_H(i, i - 1) = m_fac_H(i - 1, i);
                 m_fac_H(i, i) += Vf[i];
                 // beta <- ||f||
-                beta = m_fac_f.norm();
+                beta = norm(m_fac_f);
 
-                Vf.noalias() = V.transpose() * m_fac_f;
+                Vf.head(i1).noalias() = inner_product(V, m_fac_f);
+                ortho_err = Vf.head(i1).cwiseAbs().maxCoeff();
                 count++;
             }
         }
     }
 
-    // Implicitly restarted Arnoldi factorization
+    // Implicitly restarted Lanczos factorization
     void restart(int k)
     {
         if(k >= m_ncv)
@@ -299,48 +325,47 @@ private:
             // H -> Q'HQ
             // Since QR = H - mu * I, we have H = QR + mu * I
             // and therefore Q'HQ = RQ + mu * I
-            m_fac_H = decomp.matrix_RQ();
+            decomp.matrix_RQ(m_fac_H);
             m_fac_H.diagonal().array() += m_ritz_val[i];
         }
         // V -> VQ, only need to update the first k+1 columns
         // Q has some elements being zero
         // The first (ncv - k + i) elements of the i-th column of Q are non-zero
         Matrix Vs(m_n, k + 1);
-        int nnz;
         for(int i = 0; i < k; i++)
         {
-            nnz = m_ncv - k + i + 1;
-            MapMat V(m_fac_V.data(), m_n, nnz);
+            const int nnz = m_ncv - k + i + 1;
             MapVec q(&Q(0, i), nnz);
-            Vs.col(i).noalias() = V * q;
+            Vs.col(i).noalias() = m_fac_V.leftCols(nnz) * q;
         }
         Vs.col(k).noalias() = m_fac_V * Q.col(k);
         m_fac_V.leftCols(k + 1).noalias() = Vs;
 
-        Vector fk = m_fac_f * Q(m_ncv - 1, k - 1) + m_fac_V.col(k) * m_fac_H(k, k - 1);
+        const Vector fk = m_fac_f * Q(m_ncv - 1, k - 1) + m_fac_V.col(k) * m_fac_H(k, k - 1);
         factorize_from(k, m_ncv, fk);
         retrieve_ritzpair();
     }
 
-    // Calculate the number of converged Ritz values
+    // Calculates the number of converged Ritz values
     int num_converged(Scalar tol)
     {
-        // thresh = tol * max(m_prec, abs(theta)), theta for ritz value
-        Array thresh = tol * m_ritz_val.head(m_nev).array().abs().max(m_prec);
-        Array resid =  m_ritz_est.head(m_nev).array().abs() * m_fac_f.norm();
-        // Converged "wanted" ritz values
+        // thresh = tol * max(m_eps23, abs(theta)), theta for Ritz value
+        Array thresh = tol * m_ritz_val.head(m_nev).array().abs().max(m_eps23);
+        Array resid =  m_ritz_est.head(m_nev).array().abs() * norm(m_fac_f);
+        // Converged "wanted" Ritz values
         m_ritz_conv = (resid < thresh);
 
         return m_ritz_conv.cast<int>().sum();
     }
 
-    // Return the adjusted nev for restarting
+    // Returns the adjusted nev for restarting
     int nev_adjusted(int nconv)
     {
-        int nev_new = m_nev;
+        using std::abs;
 
+        int nev_new = m_nev;
         for(int i = m_nev; i < m_ncv; i++)
-            if(std::abs(m_ritz_est[i]) < m_prec)  nev_new++;
+            if(abs(m_ritz_est[i]) < m_near_0)  nev_new++;
 
         // Adjust nev_new, according to dsaup2.f line 677~684 in ARPACK
         nev_new += std::min(nconv, (m_ncv - nev_new) / 2);
@@ -349,15 +374,18 @@ private:
         else if(nev_new == 1 && m_ncv > 2)
             nev_new = 2;
 
+        if(nev_new > m_ncv - 1)
+            nev_new = m_ncv - 1;
+
         return nev_new;
     }
 
-    // Retrieve and sort ritz values and ritz vectors
+    // Retrieves and sorts Ritz values and Ritz vectors
     void retrieve_ritzpair()
     {
         TridiagEigen<Scalar> decomp(m_fac_H);
-        Vector evals = decomp.eigenvalues();
-        Matrix evecs = decomp.eigenvectors();
+        const Vector& evals = decomp.eigenvalues();
+        const Matrix& evecs = decomp.eigenvectors();
 
         SortEigenvalue<Scalar, SelectionRule> sorting(evals.data(), evals.size());
         std::vector<int> ind = sorting.index();
@@ -384,7 +412,7 @@ private:
             }
         }
 
-        // Copy the ritz values and vectors to m_ritz_val and m_ritz_vec, respectively
+        // Copy the Ritz values and vectors to m_ritz_val and m_ritz_vec, respectively
         for(int i = 0; i < m_ncv; i++)
         {
             m_ritz_val[i] = evals[ind[i]];
@@ -392,12 +420,21 @@ private:
         }
         for(int i = 0; i < m_nev; i++)
         {
-            m_ritz_vec.col(i) = evecs.col(ind[i]);
+            m_ritz_vec.col(i).noalias() = evecs.col(ind[i]);
         }
     }
 
 protected:
-    // Sort the first nev Ritz pairs in the specified order
+    // In generalized eigenvalue problem Ax=lambda*Bx, define the inner product to be <x, y> = x'By
+    // For regular eigenvalue problems, it is the usual inner product <x, y> = x'y
+    virtual Scalar inner_product(const Vector& x, const Vector& y) { return x.dot(y); }
+    virtual Scalar inner_product(const MapVec& x, const Vector& y) { return x.dot(y); }
+    virtual Vector inner_product(const MapMat& x, const Vector& y) { return x.transpose() * y; }
+
+    // B-norm of a vector. For regular eigenvalue problems it is simply the L2 norm
+    virtual Scalar norm(const Vector& x) { return x.norm(); }
+
+    // Sorts the first nev Ritz pairs in the specified order
     // This is used to return the final results
     virtual void sort_ritzpair(int sort_rule)
     {
@@ -438,7 +475,7 @@ protected:
         for(int i = 0; i < m_nev; i++)
         {
             new_ritz_val[i] = m_ritz_val[ind[i]];
-            new_ritz_vec.col(i) = m_ritz_vec.col(ind[i]);
+            new_ritz_vec.col(i).noalias() = m_ritz_vec.col(ind[i]);
             new_ritz_conv[i] = m_ritz_conv[ind[i]];
         }
 
@@ -451,39 +488,46 @@ public:
     ///
     /// Constructor to create a solver object.
     ///
-    /// \param op_  Pointer to the matrix operation object, which should implement
+    /// \param op   Pointer to the matrix operation object, which should implement
     ///             the matrix-vector multiplication operation of \f$A\f$:
-    ///             calculating \f$Ay\f$ for any vector \f$y\f$. Users could either
-    ///             create the object from the DenseSymMatProd wrapper class, or
-    ///             define their own that impelemnts all the public member functions
+    ///             calculating \f$Av\f$ for any vector \f$v\f$. Users could either
+    ///             create the object from the wrapper class such as DenseSymMatProd, or
+    ///             define their own that impelements all the public member functions
     ///             as in DenseSymMatProd.
-    /// \param nev_ Number of eigenvalues requested. This should satisfy \f$1\le nev \le n-1\f$,
+    /// \param nev  Number of eigenvalues requested. This should satisfy \f$1\le nev \le n-1\f$,
     ///             where \f$n\f$ is the size of matrix.
-    /// \param ncv_ Parameter that controls the convergence speed of the algorithm.
-    ///             Typically a larger `ncv_` means faster convergence, but it may
+    /// \param ncv  Parameter that controls the convergence speed of the algorithm.
+    ///             Typically a larger `ncv` means faster convergence, but it may
     ///             also result in greater memory use and more matrix operations
     ///             in each iteration. This parameter must satisfy \f$nev < ncv \le n\f$,
     ///             and is advised to take \f$ncv \ge 2\cdot nev\f$.
     ///
-    SymEigsSolver(OpType *op_, int nev_, int ncv_) :
-        m_op(op_),
+    SymEigsSolver(OpType* op, int nev, int ncv) :
+        m_op(op),
         m_n(m_op->rows()),
-        m_nev(nev_),
-        m_ncv(ncv_ > m_n ? m_n : ncv_),
+        m_nev(nev),
+        m_ncv(ncv > m_n ? m_n : ncv),
         m_nmatop(0),
         m_niter(0),
         m_info(NOT_COMPUTED),
-        m_prec(std::pow(std::numeric_limits<Scalar>::epsilon(), Scalar(2.0) / 3))
+        m_near_0(TypeTraits<Scalar>::min() * Scalar(10)),
+        m_eps(Eigen::NumTraits<Scalar>::epsilon()),
+        m_eps23(Eigen::numext::pow(m_eps, Scalar(2.0) / 3))
     {
-        if(nev_ < 1 || nev_ > m_n - 1)
+        if(nev < 1 || nev > m_n - 1)
             throw std::invalid_argument("nev must satisfy 1 <= nev <= n - 1, n is the size of matrix");
 
-        if(ncv_ <= nev_ || ncv_ > m_n)
+        if(ncv <= nev || ncv > m_n)
             throw std::invalid_argument("ncv must satisfy nev < ncv <= n, n is the size of matrix");
     }
 
     ///
-    /// Providing the initial residual vector for the algorithm.
+    /// Virtual destructor
+    ///
+    virtual ~SymEigsSolver() {}
+
+    ///
+    /// Initializes the solver by providing an initial residual vector.
     ///
     /// \param init_resid Pointer to the initial residual vector.
     ///
@@ -491,7 +535,7 @@ public:
     /// to find eigenvalues. This function allows the user to provide the initial
     /// residual vector.
     ///
-    void init(const Scalar *init_resid)
+    void init(const Scalar* init_resid)
     {
         // Reset all matrices/vectors to zero
         m_fac_V.resize(m_n, m_ncv);
@@ -513,10 +557,11 @@ public:
         m_nmatop = 0;
         m_niter = 0;
 
+        // Set the initial vector
         Vector v(m_n);
         std::copy(init_resid, init_resid + m_n, v.data());
-        Scalar vnorm = v.norm();
-        if(vnorm < m_prec)
+        const Scalar vnorm = norm(v);
+        if(vnorm < m_near_0)
             throw std::invalid_argument("initial residual vector cannot be zero");
         v /= vnorm;
 
@@ -524,17 +569,22 @@ public:
         m_op->perform_op(v.data(), w.data());
         m_nmatop++;
 
-        m_fac_H(0, 0) = v.dot(w);
-        m_fac_f = w - v * m_fac_H(0, 0);
-        m_fac_V.col(0) = v;
+        m_fac_H(0, 0) = inner_product(v, w);
+        m_fac_f.noalias() = w - v * m_fac_H(0, 0);
+        m_fac_V.col(0).noalias() = v;
+
+        // In some cases f is zero in exact arithmetics, but due to rounding errors
+        // it may contain tiny fluctuations. When this happens, we force f to be zero.
+        if(m_fac_f.cwiseAbs().maxCoeff() < m_eps)
+            m_fac_f.setZero();
     }
 
     ///
-    /// Providing a random initial residual vector.
+    /// Initializes the solver by providing a random initial residual vector.
     ///
     /// This overloaded function generates a random initial residual vector
-    /// for the algorithm. Elements in the vector follow independent Uniform(-0.5, 0.5)
-    /// distributions.
+    /// (with a fixed random seed) for the algorithm. Elements in the vector
+    /// follow independent Uniform(-0.5, 0.5) distribution.
     ///
     void init()
     {
@@ -544,7 +594,7 @@ public:
     }
 
     ///
-    /// Conducting the major computation procedure.
+    /// Conducts the major computation procedure.
     ///
     /// \param maxit      Maximum number of iterations allowed in the algorithm.
     /// \param tol        Precision parameter for the calculated eigenvalues.
@@ -587,31 +637,31 @@ public:
     }
 
     ///
-    /// Retuning the status of the computation.
+    /// Returns the status of the computation.
     /// The full list of enumeration values can be found in \ref Enumerations.
     ///
-    int info() { return m_info; }
+    int info() const { return m_info; }
 
     ///
-    /// Returning the number of iterations used in the computation.
+    /// Returns the number of iterations used in the computation.
     ///
-    int num_iterations() { return m_niter; }
+    int num_iterations() const { return m_niter; }
 
     ///
-    /// Returning the number of matrix operations used in the computation.
+    /// Returns the number of matrix operations used in the computation.
     ///
-    int num_operations() { return m_nmatop; }
+    int num_operations() const { return m_nmatop; }
 
     ///
-    /// Returning the converged eigenvalues.
+    /// Returns the converged eigenvalues.
     ///
     /// \return A vector containing the eigenvalues.
     /// Returned vector type will be `Eigen::Vector<Scalar, ...>`, depending on
     /// the template parameter `Scalar` defined.
     ///
-    Vector eigenvalues()
+    Vector eigenvalues() const
     {
-        int nconv = m_ritz_conv.cast<int>().sum();
+        const int nconv = m_ritz_conv.cast<int>().sum();
         Vector res(nconv);
 
         if(!nconv)
@@ -631,7 +681,7 @@ public:
     }
 
     ///
-    /// Returning the eigenvectors associated with the converged eigenvalues.
+    /// Returns the eigenvectors associated with the converged eigenvalues.
     ///
     /// \param nvec The number of eigenvectors to return.
     ///
@@ -639,9 +689,9 @@ public:
     /// Returned matrix type will be `Eigen::Matrix<Scalar, ...>`,
     /// depending on the template parameter `Scalar` defined.
     ///
-    Matrix eigenvectors(int nvec)
+    virtual Matrix eigenvectors(int nvec) const
     {
-        int nconv = m_ritz_conv.cast<int>().sum();
+        const int nconv = m_ritz_conv.cast<int>().sum();
         nvec = std::min(nvec, nconv);
         Matrix res(m_n, nvec);
 
@@ -654,7 +704,7 @@ public:
         {
             if(m_ritz_conv[i])
             {
-                ritz_vec_conv.col(j) = m_ritz_vec.col(i);
+                ritz_vec_conv.col(j).noalias() = m_ritz_vec.col(i);
                 j++;
             }
         }
@@ -665,198 +715,15 @@ public:
     }
 
     ///
-    /// Returning all converged eigenvectors.
+    /// Returns all converged eigenvectors.
     ///
-    Matrix eigenvectors()
+    virtual Matrix eigenvectors() const
     {
         return eigenvectors(m_nev);
     }
 };
 
 
-
-
-
-///
-/// \ingroup EigenSolver
-///
-/// This class implements the eigen solver for real symmetric matrices using
-/// the **shift-and-invert mode**. The background information of the symmetric
-/// eigen solver is documented in the SymEigsSolver class. Here we focus on
-/// explaining the shift-and-invert mode.
-///
-/// The shift-and-invert mode is based on the following fact:
-/// If \f$\lambda\f$ and \f$x\f$ are a pair of eigenvalue and eigenvector of
-/// matrix \f$A\f$, such that \f$Ax=\lambda x\f$, then for any \f$\sigma\f$,
-/// we have
-/// \f[(A-\sigma I)^{-1}x=\nu x\f]
-/// where
-/// \f[\nu=\frac{1}{\lambda-\sigma}\f]
-/// which indicates that \f$(\nu, x)\f$ is an eigenpair of the matrix
-/// \f$(A-\sigma I)^{-1}\f$.
-///
-/// Therefore, if we pass the matrix operation \f$(A-\sigma I)^{-1}y\f$
-/// (rather than \f$Ay\f$) to the eigen solver, then we would get the desired
-/// values of \f$\nu\f$, and \f$\lambda\f$ can also be easily obtained by noting
-/// that \f$\lambda=\sigma+\nu^{-1}\f$.
-///
-/// The reason why we need this type of manipulation is that
-/// the algorithm of **Spectra** (and also **ARPACK**)
-/// is good at finding eigenvalues with large magnitude, but may fail in looking
-/// for eigenvalues that are close to zero. However, if we really need them, we
-/// can set \f$\sigma=0\f$, find the largest eigenvalues of \f$A^{-1}\f$, and then
-/// transform back to \f$\lambda\f$, since in this case largest values of \f$\nu\f$
-/// implies smallest values of \f$\lambda\f$.
-///
-/// To summarize, in the shift-and-invert mode, the selection rule will apply to
-/// \f$\nu=1/(\lambda-\sigma)\f$ rather than \f$\lambda\f$. So a selection rule
-/// of `LARGEST_MAGN` combined with shift \f$\sigma\f$ will find eigenvalues of
-/// \f$A\f$ that are closest to \f$\sigma\f$. But note that the eigenvalues()
-/// method will always return the eigenvalues in the original problem (i.e.,
-/// returning \f$\lambda\f$ rather than \f$\nu\f$), and eigenvectors are the
-/// same for both the original problem and the shifted-and-inverted problem.
-///
-/// \tparam Scalar        The element type of the matrix.
-///                       Currently supported types are `float`, `double` and `long double`.
-/// \tparam SelectionRule An enumeration value indicating the selection rule of
-///                       the shifted-and-inverted eigenvalues.
-///                       The full list of enumeration values can be found in
-///                       \ref Enumerations.
-/// \tparam OpType        The name of the matrix operation class. Users could either
-///                       use the DenseSymShiftSolve wrapper class, or define their
-///                       own that impelemnts all the public member functions as in
-///                       DenseSymShiftSolve.
-///
-/// Below is an example that illustrates the use of the shift-and-invert mode:
-///
-/// \code{.cpp}
-/// #include <Eigen/Core>
-/// #include <SymEigsSolver.h>  // Also includes <MatOp/DenseSymShiftSolve.h>
-/// #include <iostream>
-///
-/// using namespace Spectra;
-///
-/// int main()
-/// {
-///     // A size-10 diagonal matrix with elements 1, 2, ..., 10
-///     Eigen::MatrixXd M = Eigen::MatrixXd::Zero(10, 10);
-///     for(int i = 0; i < M.rows(); i++)
-///         M(i, i) = i + 1;
-///
-///     // Construct matrix operation object using the wrapper class
-///     DenseSymShiftSolve<double> op(M);
-///
-///     // Construct eigen solver object with shift 0
-///     // This will find eigenvalues that are closest to 0
-///     SymEigsShiftSolver< double, LARGEST_MAGN,
-///                         DenseSymShiftSolve<double> > eigs(&op, 3, 6, 0.0);
-///
-///     eigs.init();
-///     eigs.compute();
-///     if(eigs.info() == SUCCESSFUL)
-///     {
-///         Eigen::VectorXd evalues = eigs.eigenvalues();
-///         // Will get (3.0, 2.0, 1.0)
-///         std::cout << "Eigenvalues found:\n" << evalues << std::endl;
-///     }
-///
-///     return 0;
-/// }
-/// \endcode
-///
-/// Also an example for user-supplied matrix shift-solve operation class:
-///
-/// \code{.cpp}
-/// #include <Eigen/Core>
-/// #include <SymEigsSolver.h>
-/// #include <iostream>
-///
-/// using namespace Spectra;
-///
-/// // M = diag(1, 2, ..., 10)
-/// class MyDiagonalTenShiftSolve
-/// {
-/// private:
-///     double sigma_;
-/// public:
-///     int rows() { return 10; }
-///     int cols() { return 10; }
-///     void set_shift(double sigma) { sigma_ = sigma; }
-///     // y_out = inv(A - sigma * I) * x_in
-///     // inv(A - sigma * I) = diag(1/(1-sigma), 1/(2-sigma), ...)
-///     void perform_op(double* x_in, double* y_out)
-///     {
-///         for(int i = 0; i < rows(); i++)
-///         {
-///             y_out[i] = x_in[i] / (i + 1 - sigma_);
-///         }
-///     }
-/// };
-///
-/// int main()
-/// {
-///     MyDiagonalTenShiftSolve op;
-///     // Find three eigenvalues that are closest to 3.14
-///     SymEigsShiftSolver<double, LARGEST_MAGN,
-///                        MyDiagonalTenShiftSolve> eigs(&op, 3, 6, 3.14);
-///     eigs.init();
-///     eigs.compute();
-///     if(eigs.info() == SUCCESSFUL)
-///     {
-///         Eigen::VectorXd evalues = eigs.eigenvalues();
-///         // Will get (4.0, 3.0, 2.0)
-///         std::cout << "Eigenvalues found:\n" << evalues << std::endl;
-///     }
-///
-///     return 0;
-/// }
-/// \endcode
-///
-template <typename Scalar = double,
-          int SelectionRule = LARGEST_MAGN,
-          typename OpType = DenseSymShiftSolve<double> >
-class SymEigsShiftSolver: public SymEigsSolver<Scalar, SelectionRule, OpType>
-{
-private:
-    typedef Eigen::Array<Scalar, Eigen::Dynamic, 1> Array;
-
-    Scalar sigma;
-
-    // First transform back the ritz values, and then sort
-    void sort_ritzpair(int sort_rule)
-    {
-        Array m_ritz_val_org = Scalar(1.0) / this->m_ritz_val.head(this->m_nev).array() + sigma;
-        this->m_ritz_val.head(this->m_nev) = m_ritz_val_org;
-        SymEigsSolver<Scalar, SelectionRule, OpType>::sort_ritzpair(sort_rule);
-    }
-public:
-    ///
-    /// Constructor to create a eigen solver object using the shift-and-invert mode.
-    ///
-    /// \param op_    Pointer to the matrix operation object, which should implement
-    ///               the shift-solve operation of \f$A\f$: calculating
-    ///               \f$(A-\sigma I)^{-1}y\f$ for any vector \f$y\f$. Users could either
-    ///               create the object from the DenseSymShiftSolve wrapper class, or
-    ///               define their own that impelemnts all the public member functions
-    ///               as in DenseSymShiftSolve.
-    /// \param nev_   Number of eigenvalues requested. This should satisfy \f$1\le nev \le n-1\f$,
-    ///               where \f$n\f$ is the size of matrix.
-    /// \param ncv_   Parameter that controls the convergence speed of the algorithm.
-    ///               Typically a larger `ncv_` means faster convergence, but it may
-    ///               also result in greater memory use and more matrix operations
-    ///               in each iteration. This parameter must satisfy \f$nev < ncv \le n\f$,
-    ///               and is advised to take \f$ncv \ge 2\cdot nev\f$.
-    /// \param sigma_ The value of the shift.
-    ///
-    SymEigsShiftSolver(OpType *op_, int nev_, int ncv_, Scalar sigma_) :
-        SymEigsSolver<Scalar, SelectionRule, OpType>(op_, nev_, ncv_),
-        sigma(sigma_)
-    {
-        this->m_op->set_shift(sigma);
-    }
-};
-
-
 } // namespace Spectra
 
 #endif // SYM_EIGS_SOLVER_H
